from sqlframe.duckdb.catalog import DuckDBCatalog
from sqlframe.duckdb.column import DuckDBColumn
from sqlframe.duckdb.dataframe import DuckDBDataFrame, DuckDBDataFrameNaFunctions
from sqlframe.duckdb.group import DuckDBGroupedData
from sqlframe.duckdb.readwriter import DuckDBDataFrameReader, DuckDBDataFrameWriter
from sqlframe.duckdb.session import DuckDBSession
from sqlframe.duckdb.window import Window, WindowSpec

__all__ = [
    "DuckDBCatalog",
    "DuckDBColumn",
    "DuckDBDataFrame",
    "DuckDBDataFrameNaFunctions",
    "DuckDBGroupedData",
    "DuckDBDataFrameReader",
    "DuckDBDataFrameWriter",
    "DuckDBSession",
    "Window",
    "WindowSpec",
]



from sqlframe.snowflake.catalog import SnowflakeCatalog
from sqlframe.snowflake.column import Column
from sqlframe.snowflake.dataframe import (
    SnowflakeDataFrame,
    SnowflakeDataFrameNaFunctions,
)
from sqlframe.snowflake.group import SnowflakeGroupedData
from sqlframe.snowflake.readwriter import (
    SnowflakeDataFrameReader,
    SnowflakeDataFrameWriter,
)
from sqlframe.snowflake.session import SnowflakeSession
from sqlframe.snowflake.window import Window, WindowSpec

__all__ = [
    "SnowflakeCatalog",
    "Column",
    "SnowflakeDataFrame",
    "SnowflakeDataFrameNaFunctions",
    "SnowflakeGroupedData",
    "SnowflakeDataFrameReader",
    "SnowflakeDataFrameWriter",
    "SnowflakeSession",
    "Window",
    "WindowSpec",
]



from __future__ import annotations

import json
import typing as t
import warnings

from sqlframe.snowflake.udf import SnowflakeUDFRegistration

try:
    from snowflake.connector.converter import SnowflakeConverter
except ImportError:
    SnowflakeConverter = object  # type: ignore

from sqlframe.base.session import _BaseSession
from sqlframe.snowflake.catalog import SnowflakeCatalog
from sqlframe.snowflake.dataframe import SnowflakeDataFrame
from sqlframe.snowflake.readwriter import (
    SnowflakeDataFrameReader,
    SnowflakeDataFrameWriter,
)

if t.TYPE_CHECKING:
    from snowflake.connector import SnowflakeConnection
else:
    SnowflakeConnection = t.Any


class JsonLoadsSnowflakeConverter(SnowflakeConverter):
    # This might not be needed once proper arrow types are supported.
    # Checkout this PR: https://github.com/snowflakedb/snowflake-connector-python/pull/1853/files
    # Specifically see if `alter session set enable_structured_types_in_client_response=true` and
    # `alter session set force_enable_structured_types_native_arrow_format=true` are supported then it might work.
    # At the time of writing these were not supported parameters on my version on Snowflake.
    def _json_loads(self, ctx: dict[str, t.Any]) -> t.Callable:
        def conv(value: str) -> t.List:
            # Snowflake returns "undefined" for null values when inside an array
            # We check if we replaced "'undefined'" string and if so we switch it back
            # this is a lazy approach compared to writing a proper regex replace
            return json.loads(value.replace("undefined", "null").replace("'null'", "undefined"))

        return conv

    _OBJECT_to_python = _json_loads  # type: ignore
    _VARIANT_to_python = _json_loads  # type: ignore
    _ARRAY_to_python = _json_loads  # type: ignore


class SnowflakeSession(
    _BaseSession[  # type: ignore
        SnowflakeCatalog,
        SnowflakeDataFrameReader,
        SnowflakeDataFrameWriter,
        SnowflakeDataFrame,
        SnowflakeConnection,
        SnowflakeUDFRegistration,
    ],
):
    _catalog = SnowflakeCatalog
    _reader = SnowflakeDataFrameReader
    _writer = SnowflakeDataFrameWriter
    _df = SnowflakeDataFrame
    _udf_registration = SnowflakeUDFRegistration

    def __init__(self, conn: t.Optional[SnowflakeConnection] = None):
        import snowflake

        snowflake.connector.cursor.CAN_USE_ARROW_RESULT_FORMAT = False

        if not hasattr(self, "_conn"):
            super().__init__(conn)
            if self._conn.converter and not isinstance(
                self._conn.converter, JsonLoadsSnowflakeConverter
            ):
                self._conn.converter = JsonLoadsSnowflakeConverter(
                    use_numpy=self._conn._numpy,
                    support_negative_year=self._conn._support_negative_year,
                )
            else:
                self._conn._converter_class = JsonLoadsSnowflakeConverter  # type: ignore

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "snowflake"

        @property
        def session(self) -> SnowflakeSession:
            return SnowflakeSession(**self._session_kwargs)

        def getOrCreate(self) -> SnowflakeSession:
            self._set_session_properties()
            return self.session

    builder = Builder()



from __future__ import annotations

import math
import typing as t
import warnings

import pytest
from _pytest.fixtures import FixtureRequest
from pyspark.sql import DataFrame as PySparkDataFrame
from pyspark.sql import SparkSession as PySparkSession
from pyspark.sql import types as PySparkTypes
from sqlglot import Dialect
from sqlglot.helper import ensure_list

from sqlframe.base.types import Row
from sqlframe.base.util import (
    dialect_to_string,
)
from sqlframe.base.util import (
    get_func_from_session as get_func_from_session_without_fallback,
)
from sqlframe.bigquery import types as BigQueryTypes
from sqlframe.bigquery.dataframe import BigQueryDataFrame
from sqlframe.bigquery.session import BigQuerySession
from sqlframe.duckdb import types as DuckDBTypes
from sqlframe.duckdb.dataframe import DuckDBDataFrame
from sqlframe.duckdb.session import DuckDBSession
from sqlframe.postgres import types as PostgresTypes
from sqlframe.postgres.dataframe import PostgresDataFrame
from sqlframe.postgres.session import PostgresSession
from sqlframe.redshift import types as RedshiftTypes
from sqlframe.redshift.dataframe import RedshiftDataFrame
from sqlframe.redshift.session import RedshiftSession
from sqlframe.snowflake import types as SnowflakeTypes
from sqlframe.snowflake.dataframe import SnowflakeDataFrame
from sqlframe.snowflake.session import SnowflakeSession
from sqlframe.spark import types as SparkTypes
from sqlframe.spark.dataframe import SparkDataFrame
from sqlframe.spark.session import SparkSession
from sqlframe.standalone import types as StandaloneTypes
from sqlframe.standalone.dataframe import StandaloneDataFrame
from sqlframe.standalone.session import StandaloneSession

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import _BaseDataFrame
    from sqlframe.base.session import _BaseSession
    from tests.types import DistrictData, EmployeeData, StoreData

pytest_plugins = ["tests.common_fixtures"]


ENGINE_PARAMETERS = [
    pytest.param(
        "pyspark",
        marks=[
            pytest.mark.pyspark_session,
            pytest.mark.local,
        ],
    ),
    pytest.param(
        "standalone",
        marks=[
            pytest.mark.standalone,
            pytest.mark.local,
        ],
    ),
    pytest.param(
        "duckdb",
        marks=[
            pytest.mark.duckdb,
            pytest.mark.local,
        ],
    ),
    pytest.param(
        "postgres",
        marks=[
            pytest.mark.postgres,
            pytest.mark.local,
        ],
    ),
    pytest.param(
        "bigquery",
        marks=[
            pytest.mark.bigquery,
            pytest.mark.remote,
            # Set xdist group in order to serialize tests
            pytest.mark.xdist_group("bigquery_tests"),
        ],
    ),
    pytest.param(
        "snowflake",
        marks=[
            pytest.mark.snowflake,
            pytest.mark.remote,
            # Set xdist group in order to serialize tests
            pytest.mark.xdist_group("snowflake_tests"),
        ],
    ),
    pytest.param(
        "spark",
        marks=[
            pytest.mark.spark_session,
            pytest.mark.local,
        ],
    ),
    # pytest.param(
    #     "redshift",
    #     marks=[
    #         pytest.mark.redshift,
    #         pytest.mark.remote,
    #         # Set xdist group in order to serialize tests
    #         pytest.mark.xdist_group("redshift_tests"),
    #     ],
    # ),
]

ENGINE_PARAMETERS_NO_PYSPARK_STANDALONE = [
    x for x in ENGINE_PARAMETERS if "pyspark" not in x.values and "standalone" not in x.values
]

ENGINE_PARAMETERS_NO_STANDALONE = [x for x in ENGINE_PARAMETERS if "standalone" not in x.values]

ENGINE_PARAMETERS_NO_PYSPARK = [x for x in ENGINE_PARAMETERS if "pyspark" not in x.values]


@pytest.fixture(scope="module", autouse=True)
def filter_warnings():
    warnings.filterwarnings("ignore", category=ResourceWarning)


@pytest.fixture(scope="module")
def _employee_data() -> EmployeeData:
    return [
        (1, "Jack", "Shephard", 37, 1),
        (2, "John", "Locke", 65, 1),
        (3, "Kate", "Austen", 37, 2),
        (4, "Claire", "Littleton", 27, 2),
        (5, "Hugo", "Reyes", 29, 100),
    ]


@pytest.fixture(scope="module")
def _store_data() -> StoreData:
    return [
        (1, "Hydra", 1, 37),
        (2, "Arrow", 2, 2000),
    ]


@pytest.fixture(scope="module")
def _district_data() -> t.List[t.Tuple[int, str]]:
    return [
        (1, "Los Angeles"),
        (2, "Santa Monica"),
    ]


@pytest.fixture(scope="module")
def pyspark_employee(
    pyspark_session: PySparkSession, _employee_data: EmployeeData
) -> PySparkDataFrame:
    pyspark_employee_schema = PySparkTypes.StructType(
        [
            PySparkTypes.StructField("employee_id", PySparkTypes.IntegerType(), False),
            PySparkTypes.StructField("fname", PySparkTypes.StringType(), False),
            PySparkTypes.StructField("lname", PySparkTypes.StringType(), False),
            PySparkTypes.StructField("age", PySparkTypes.IntegerType(), False),
            PySparkTypes.StructField("store_id", PySparkTypes.IntegerType(), False),
        ]
    )
    df = pyspark_session.createDataFrame(data=_employee_data, schema=pyspark_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture(scope="module")
def pyspark_store(pyspark_session: PySparkSession, _store_data: StoreData) -> PySparkDataFrame:
    spark_store_schema = PySparkTypes.StructType(
        [
            PySparkTypes.StructField("store_id", PySparkTypes.IntegerType(), False),
            PySparkTypes.StructField("store_name", PySparkTypes.StringType(), False),
            PySparkTypes.StructField("district_id", PySparkTypes.IntegerType(), False),
            PySparkTypes.StructField("num_sales", PySparkTypes.IntegerType(), False),
        ]
    )
    df = pyspark_session.createDataFrame(data=_store_data, schema=spark_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture(scope="module")
def pyspark_district(
    pyspark_session: PySparkSession, _district_data: DistrictData
) -> PySparkDataFrame:
    spark_district_schema = PySparkTypes.StructType(
        [
            PySparkTypes.StructField("district_id", PySparkTypes.IntegerType(), False),
            PySparkTypes.StructField("district_name", PySparkTypes.StringType(), False),
        ]
    )
    df = pyspark_session.createDataFrame(data=_district_data, schema=spark_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture(scope="function")
def standalone_store(
    standalone_session: StandaloneSession, _store_data: StoreData
) -> StandaloneDataFrame:
    sqlf_store_schema = StandaloneTypes.StructType(
        [
            StandaloneTypes.StructField("store_id", StandaloneTypes.IntegerType(), False),
            StandaloneTypes.StructField("store_name", StandaloneTypes.StringType(), False),
            StandaloneTypes.StructField("district_id", StandaloneTypes.IntegerType(), False),
            StandaloneTypes.StructField("num_sales", StandaloneTypes.IntegerType(), False),
        ]
    )
    df = standalone_session.createDataFrame(data=_store_data, schema=sqlf_store_schema)
    standalone_session.catalog.add_table("store", sqlf_store_schema)  # type: ignore
    return df


@pytest.fixture(scope="function")
def standalone_district(
    standalone_session: StandaloneSession, _district_data: DistrictData
) -> StandaloneDataFrame:
    sqlf_district_schema = StandaloneTypes.StructType(
        [
            StandaloneTypes.StructField("district_id", StandaloneTypes.IntegerType(), False),
            StandaloneTypes.StructField("district_name", StandaloneTypes.StringType(), False),
        ]
    )
    df = standalone_session.createDataFrame(data=_district_data, schema=sqlf_district_schema)
    standalone_session.catalog.add_table("district", sqlf_district_schema)  # type: ignore
    return df


@pytest.fixture(scope="function")
def duckdb_employee(duckdb_session: DuckDBSession, _employee_data: EmployeeData) -> DuckDBDataFrame:
    duckdb_employee_schema = DuckDBTypes.StructType(
        [
            DuckDBTypes.StructField("employee_id", DuckDBTypes.IntegerType(), False),
            DuckDBTypes.StructField("fname", DuckDBTypes.StringType(), False),
            DuckDBTypes.StructField("lname", DuckDBTypes.StringType(), False),
            DuckDBTypes.StructField("age", DuckDBTypes.IntegerType(), False),
            DuckDBTypes.StructField("store_id", DuckDBTypes.IntegerType(), False),
        ]
    )
    df = duckdb_session.createDataFrame(data=_employee_data, schema=duckdb_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture(scope="function")
def duckdb_store(duckdb_session: DuckDBSession, _store_data: StoreData) -> DuckDBDataFrame:
    duckdb_store_schema = DuckDBTypes.StructType(
        [
            DuckDBTypes.StructField("store_id", DuckDBTypes.IntegerType(), False),
            DuckDBTypes.StructField("store_name", DuckDBTypes.StringType(), False),
            DuckDBTypes.StructField("district_id", DuckDBTypes.IntegerType(), False),
            DuckDBTypes.StructField("num_sales", DuckDBTypes.IntegerType(), False),
        ]
    )
    df = duckdb_session.createDataFrame(data=_store_data, schema=duckdb_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture(scope="function")
def duckdb_district(duckdb_session: DuckDBSession, _district_data: DistrictData) -> DuckDBDataFrame:
    duckdb_district_schema = DuckDBTypes.StructType(
        [
            DuckDBTypes.StructField("district_id", DuckDBTypes.IntegerType(), False),
            DuckDBTypes.StructField("district_name", DuckDBTypes.StringType(), False),
        ]
    )
    df = duckdb_session.createDataFrame(data=_district_data, schema=duckdb_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture
def postgres_employee(
    postgres_session: PostgresSession, _employee_data: EmployeeData
) -> PostgresDataFrame:
    postgres_employee_schema = PostgresTypes.StructType(
        [
            PostgresTypes.StructField("employee_id", PostgresTypes.IntegerType(), False),
            PostgresTypes.StructField("fname", PostgresTypes.StringType(), False),
            PostgresTypes.StructField("lname", PostgresTypes.StringType(), False),
            PostgresTypes.StructField("age", PostgresTypes.IntegerType(), False),
            PostgresTypes.StructField("store_id", PostgresTypes.IntegerType(), False),
        ]
    )
    df = postgres_session.createDataFrame(data=_employee_data, schema=postgres_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture
def postgres_store(postgres_session: PostgresSession, _store_data: StoreData) -> PostgresDataFrame:
    postgres_store_schema = PostgresTypes.StructType(
        [
            PostgresTypes.StructField("store_id", PostgresTypes.IntegerType(), False),
            PostgresTypes.StructField("store_name", PostgresTypes.StringType(), False),
            PostgresTypes.StructField("district_id", PostgresTypes.IntegerType(), False),
            PostgresTypes.StructField("num_sales", PostgresTypes.IntegerType(), False),
        ]
    )
    df = postgres_session.createDataFrame(data=_store_data, schema=postgres_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture
def postgres_district(
    postgres_session: PostgresSession, _district_data: DistrictData
) -> PostgresDataFrame:
    postgres_district_schema = PostgresTypes.StructType(
        [
            PostgresTypes.StructField("district_id", PostgresTypes.IntegerType(), False),
            PostgresTypes.StructField("district_name", PostgresTypes.StringType(), False),
        ]
    )
    df = postgres_session.createDataFrame(data=_district_data, schema=postgres_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture(scope="function")
def spark_employee(spark_session: SparkSession, _employee_data: EmployeeData) -> SparkDataFrame:
    spark_employee_schema = SparkTypes.StructType(
        [
            SparkTypes.StructField("employee_id", SparkTypes.IntegerType(), False),
            SparkTypes.StructField("fname", SparkTypes.StringType(), False),
            SparkTypes.StructField("lname", SparkTypes.StringType(), False),
            SparkTypes.StructField("age", SparkTypes.IntegerType(), False),
            SparkTypes.StructField("store_id", SparkTypes.IntegerType(), False),
        ]
    )
    df = spark_session.createDataFrame(data=_employee_data, schema=spark_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture(scope="function")
def spark_store(spark_session: SparkSession, _store_data: StoreData) -> SparkDataFrame:
    spark_store_schema = SparkTypes.StructType(
        [
            SparkTypes.StructField("store_id", SparkTypes.IntegerType(), False),
            SparkTypes.StructField("store_name", SparkTypes.StringType(), False),
            SparkTypes.StructField("district_id", SparkTypes.IntegerType(), False),
            SparkTypes.StructField("num_sales", SparkTypes.IntegerType(), False),
        ]
    )
    df = spark_session.createDataFrame(data=_store_data, schema=spark_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture(scope="function")
def spark_district(spark_session: SparkSession, _district_data: DistrictData) -> SparkDataFrame:
    spark_district_schema = SparkTypes.StructType(
        [
            SparkTypes.StructField("district_id", SparkTypes.IntegerType(), False),
            SparkTypes.StructField("district_name", SparkTypes.StringType(), False),
        ]
    )
    df = spark_session.createDataFrame(data=_district_data, schema=spark_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture
def bigquery_employee(
    bigquery_session: BigQuerySession, _employee_data: EmployeeData
) -> BigQueryDataFrame:
    bigquery_employee_schema = BigQueryTypes.StructType(
        [
            BigQueryTypes.StructField("employee_id", BigQueryTypes.IntegerType(), False),
            BigQueryTypes.StructField("fname", BigQueryTypes.StringType(), False),
            BigQueryTypes.StructField("lname", BigQueryTypes.StringType(), False),
            BigQueryTypes.StructField("age", BigQueryTypes.IntegerType(), False),
            BigQueryTypes.StructField("store_id", BigQueryTypes.IntegerType(), False),
        ]
    )
    df = bigquery_session.createDataFrame(data=_employee_data, schema=bigquery_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture
def bigquery_store(bigquery_session: BigQuerySession, _store_data: StoreData) -> BigQueryDataFrame:
    bigquery_store_schema = BigQueryTypes.StructType(
        [
            BigQueryTypes.StructField("store_id", BigQueryTypes.IntegerType(), False),
            BigQueryTypes.StructField("store_name", BigQueryTypes.StringType(), False),
            BigQueryTypes.StructField("district_id", BigQueryTypes.IntegerType(), False),
            BigQueryTypes.StructField("num_sales", BigQueryTypes.IntegerType(), False),
        ]
    )
    df = bigquery_session.createDataFrame(data=_store_data, schema=bigquery_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture
def bigquery_district(
    bigquery_session: BigQuerySession, _district_data: DistrictData
) -> BigQueryDataFrame:
    bigquery_district_schema = BigQueryTypes.StructType(
        [
            BigQueryTypes.StructField("district_id", BigQueryTypes.IntegerType(), False),
            BigQueryTypes.StructField("district_name", BigQueryTypes.StringType(), False),
        ]
    )
    df = bigquery_session.createDataFrame(data=_district_data, schema=bigquery_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture
def redshift_employee(
    redshift_session: RedshiftSession, _employee_data: EmployeeData
) -> RedshiftDataFrame:
    redshift_employee_schema = RedshiftTypes.StructType(
        [
            RedshiftTypes.StructField("employee_id", RedshiftTypes.IntegerType(), False),
            RedshiftTypes.StructField("fname", RedshiftTypes.StringType(), False),
            RedshiftTypes.StructField("lname", RedshiftTypes.StringType(), False),
            RedshiftTypes.StructField("age", RedshiftTypes.IntegerType(), False),
            RedshiftTypes.StructField("store_id", RedshiftTypes.IntegerType(), False),
        ]
    )
    df = redshift_session.createDataFrame(data=_employee_data, schema=redshift_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture
def redshift_store(redshift_session: RedshiftSession, _store_data: StoreData) -> RedshiftDataFrame:
    redshift_store_schema = RedshiftTypes.StructType(
        [
            RedshiftTypes.StructField("store_id", RedshiftTypes.IntegerType(), False),
            RedshiftTypes.StructField("store_name", RedshiftTypes.StringType(), False),
            RedshiftTypes.StructField("district_id", RedshiftTypes.IntegerType(), False),
            RedshiftTypes.StructField("num_sales", RedshiftTypes.IntegerType(), False),
        ]
    )
    df = redshift_session.createDataFrame(data=_store_data, schema=redshift_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture
def redshift_district(
    redshift_session: RedshiftSession, _district_data: DistrictData
) -> RedshiftDataFrame:
    redshift_district_schema = RedshiftTypes.StructType(
        [
            RedshiftTypes.StructField("district_id", RedshiftTypes.IntegerType(), False),
            RedshiftTypes.StructField("district_name", RedshiftTypes.StringType(), False),
        ]
    )
    df = redshift_session.createDataFrame(data=_district_data, schema=redshift_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture
def snowflake_employee(
    snowflake_session: SnowflakeSession, _employee_data: EmployeeData
) -> SnowflakeDataFrame:
    snowflake_employee_schema = SnowflakeTypes.StructType(
        [
            SnowflakeTypes.StructField("employee_id", SnowflakeTypes.IntegerType(), False),
            SnowflakeTypes.StructField("fname", SnowflakeTypes.StringType(), False),
            SnowflakeTypes.StructField("lname", SnowflakeTypes.StringType(), False),
            SnowflakeTypes.StructField("age", SnowflakeTypes.IntegerType(), False),
            SnowflakeTypes.StructField("store_id", SnowflakeTypes.IntegerType(), False),
        ]
    )
    df = snowflake_session.createDataFrame(data=_employee_data, schema=snowflake_employee_schema)
    df.createOrReplaceTempView("employee")
    return df


@pytest.fixture
def snowflake_store(
    snowflake_session: SnowflakeSession, _store_data: StoreData
) -> SnowflakeDataFrame:
    snowflake_store_schema = SnowflakeTypes.StructType(
        [
            SnowflakeTypes.StructField("store_id", SnowflakeTypes.IntegerType(), False),
            SnowflakeTypes.StructField("store_name", SnowflakeTypes.StringType(), False),
            SnowflakeTypes.StructField("district_id", SnowflakeTypes.IntegerType(), False),
            SnowflakeTypes.StructField("num_sales", SnowflakeTypes.IntegerType(), False),
        ]
    )
    df = snowflake_session.createDataFrame(data=_store_data, schema=snowflake_store_schema)
    df.createOrReplaceTempView("store")
    return df


@pytest.fixture
def snowflake_district(
    snowflake_session: SnowflakeSession, _district_data: DistrictData
) -> SnowflakeDataFrame:
    snowflake_district_schema = SnowflakeTypes.StructType(
        [
            SnowflakeTypes.StructField("district_id", SnowflakeTypes.IntegerType(), False),
            SnowflakeTypes.StructField("district_name", SnowflakeTypes.StringType(), False),
        ]
    )
    df = snowflake_session.createDataFrame(data=_district_data, schema=snowflake_district_schema)
    df.createOrReplaceTempView("district")
    return df


@pytest.fixture
def compare_frames(pyspark_session: PySparkSession) -> t.Callable:
    def _make_function(
        df: PySparkDataFrame,
        sqlf: StandaloneDataFrame,
        compare_schema: bool = True,
        no_empty: bool = True,
        sort: bool = False,
        optimize: t.Optional[bool] = None,
    ) -> t.Tuple[PySparkDataFrame, PySparkDataFrame]:
        from sqlframe.base.session import _BaseSession

        def compare_schemas(schema_1, schema_2):
            for schema in [schema_1, schema_2]:
                for struct_field in schema.fields:
                    struct_field.func_metadata = {}
            assert schema_1 == schema_2

        force_sort = sort
        if not sqlf.session._has_connection:
            for statement in sqlf.sql(pretty=False, optimize=False, as_list=True):
                sqlf_df = pyspark_session.sql(statement)
        else:
            # `schema` is not an implemented method
            compare_schema = False
            # We don't care if different engines produce different sorted results
            force_sort = True
            sqlf_df = sqlf  # type: ignore
        sqlf_df_results = sqlf_df.collect()  # type: ignore
        spark_df_results = df.collect()
        if isinstance(sqlf_df, _BaseSession) and sqlf_df.session.SANITIZE_COLUMN_NAMES:
            sanitized_spark_df_results = []
            for row in spark_df_results:
                sanitized_row = {
                    sqlf_df.session._sanitize_column_name(k): v for k, v in row.asDict().items()
                }
                sanitized_spark_df_results.append(Row(**sanitized_row))
            spark_df_results = sanitized_spark_df_results
        if compare_schema:
            compare_schemas(df.schema, sqlf_df.schema)
        if force_sort:
            spark_df_results = sorted(spark_df_results, key=lambda x: "_".join(map(str, x)))
            sqlf_df_results = sorted(sqlf_df_results, key=lambda x: "_".join(map(str, x)))
        assert len(spark_df_results) == len(sqlf_df_results)
        for spark_row, sqlf_row in zip(spark_df_results, sqlf_df_results):
            assert len(spark_row) == len(sqlf_row)
            for spark_value, sqlf_value in zip(spark_row, sqlf_row):
                if isinstance(spark_value, float):
                    assert math.isclose(spark_value, sqlf_value) is True
                else:
                    assert spark_value == sqlf_value
        if no_empty:
            assert len(spark_df_results) != 0
            assert len(sqlf_df_results) != 0
        return df, sqlf_df

    return _make_function


@pytest.fixture
def get_explain_plan() -> t.Callable:
    def _make_function(df: PySparkDataFrame, mode: str = "extended") -> str:
        return df._sc._jvm.PythonSQLUtils.explainString(df._jdf.queryExecution(), mode)  # type: ignore

    return _make_function


@pytest.fixture(params=ENGINE_PARAMETERS_NO_PYSPARK_STANDALONE)
def get_engine_df(request: FixtureRequest) -> t.Callable[[str], _BaseDataFrame]:
    mapping = {
        "employee": f"{request.param}_employee",
        "store": f"{request.param}_store",
        "district": f"{request.param}_district",
    }

    def _get_engine_df(name: str) -> _BaseDataFrame:
        return request.getfixturevalue(mapping[name])

    return _get_engine_df


@pytest.fixture(params=ENGINE_PARAMETERS_NO_PYSPARK_STANDALONE)
def get_session(request: FixtureRequest) -> t.Callable[[], _BaseSession]:
    def _get_session() -> _BaseSession:
        return request.getfixturevalue(f"{request.param}_session")

    return _get_session


@pytest.fixture(params=ENGINE_PARAMETERS_NO_PYSPARK)
def get_df(request: FixtureRequest) -> t.Callable[[str], _BaseDataFrame]:
    mapping = {
        "employee": f"{request.param}_employee",
        "store": f"{request.param}_store",
        "district": f"{request.param}_district",
    }

    def _get_df(name: str) -> _BaseDataFrame:
        return request.getfixturevalue(mapping[name])

    return _get_df


@pytest.fixture(params=ENGINE_PARAMETERS_NO_STANDALONE)
def get_engine_session_and_spark(
    request: FixtureRequest,
) -> t.Callable[[], t.Union[_BaseSession, PySparkSession]]:
    def _get_engine_session_and_spark() -> t.Union[_BaseSession, PySparkSession]:
        return request.getfixturevalue(f"{request.param}_session")

    return _get_engine_session_and_spark


@pytest.fixture(params=ENGINE_PARAMETERS_NO_STANDALONE)
def get_engine_df_and_pyspark(
    request: FixtureRequest,
) -> t.Callable[[str], t.Union[_BaseDataFrame, PySparkDataFrame]]:
    mapping = {
        "employee": f"{request.param}_employee",
        "store": f"{request.param}_store",
        "district": f"{request.param}_district",
    }

    def _get_engine_df_and_pyspark(name: str) -> t.Union[_BaseDataFrame, PySparkDataFrame]:
        return request.getfixturevalue(mapping[name])

    return _get_engine_df_and_pyspark


def get_func_from_session(name: str, session: t.Union[_BaseSession, PySparkSession]) -> t.Callable:
    return get_func_from_session_without_fallback(name, session, fallback=False)


@pytest.fixture
def get_func() -> t.Callable[[str, t.Union[_BaseSession, PySparkSession]], t.Callable]:
    def _get_func(name: str, session: t.Union[_BaseSession, PySparkSession]) -> t.Callable:
        try:
            return get_func_from_session(name, session)
        except AttributeError:
            if isinstance(session, PySparkSession):
                pytest.skip(f"pyspark does not support {name}")
            else:
                pytest.skip(f"{dialect_to_string(session.input_dialect)} does not support {name}")

    return _get_func


@pytest.fixture
def is_duckdb(request: FixtureRequest) -> t.Callable:
    def _is_duckdb() -> bool:
        return request.node.name.endswith("[duckdb]")

    return _is_duckdb


@pytest.fixture
def is_postgres(request: FixtureRequest) -> t.Callable:
    def _is_postgres() -> bool:
        return request.node.name.endswith("[postgres]")

    return _is_postgres


@pytest.fixture
def is_bigquery(request: FixtureRequest) -> t.Callable:
    def _is_bigquery() -> bool:
        return request.node.name.endswith("[bigquery]")

    return _is_bigquery


@pytest.fixture
def is_redshift(request: FixtureRequest) -> t.Callable:
    def _is_redshift() -> bool:
        return request.node.name.endswith("[redshift]")

    return _is_redshift


@pytest.fixture
def is_snowflake(request: FixtureRequest) -> t.Callable:
    def _is_snowflake() -> bool:
        return request.node.name.endswith("[snowflake]")

    return _is_snowflake


@pytest.fixture
def is_spark(request: FixtureRequest) -> t.Callable:
    def _is_spark() -> bool:
        return request.node.name.endswith("[spark]")

    return _is_spark



from sqlframe.base.function_alternatives import (  # noqa
    any_value_always_ignore_nulls as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    kurtosis_from_kurtosis_pop as kurtosis,
    collect_set_from_list_distinct as collect_set,
    first_always_ignore_nulls as first,
    factorial_ensure_int as factorial,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    percentile_approx_without_accuracy as percentile_approx,
    rand_no_seed as rand,
    base64_from_blob as base64,
    decode_from_blob as decode,
    format_string_with_pipes as format_string,
    overlay_from_substr as overlay,
    split_no_limit as split,
    arrays_overlap_using_intersect as arrays_overlap,
    slice_as_list_slice as slice,
    array_join_null_replacement_with_transform as array_join,
    element_at_using_brackets as element_at,
    array_remove_using_filter as array_remove,
    array_union_using_list_concat as array_union,
    array_min_from_sort as array_min,
    array_max_from_sort as array_max,
    sequence_from_generate_series as sequence,
    try_element_at_zero_based as try_element_at,
    to_unix_timestamp_include_default_format as to_unix_timestamp,
)
from sqlframe.base.functions import (
    abs as abs,
    acos as acos,
    add_months as add_months,
    approxCountDistinct as approxCountDistinct,
    approx_count_distinct as approx_count_distinct,
    array as array,
    array_contains as array_contains,
    array_distinct as array_distinct,
    array_intersect as array_intersect,
    array_position as array_position,
    array_sort as array_sort,
    asc as asc,
    asc_nulls_first as asc_nulls_first,
    asc_nulls_last as asc_nulls_last,
    ascii as ascii,
    asin as asin,
    atan as atan,
    atan2 as atan2,
    avg as avg,
    bin as bin,
    bit_length as bit_length,
    bitwiseNOT as bitwiseNOT,
    bitwise_not as bitwise_not,
    bool_and as bool_and,
    bool_or as bool_or,
    call_function as call_function,
    cbrt as cbrt,
    ceil as ceil,
    ceiling as ceiling,
    char as char,
    coalesce as coalesce,
    col as col,
    collect_list as collect_list,
    concat as concat,
    concat_ws as concat_ws,
    corr as corr,
    cos as cos,
    cot as cot,
    count as count,
    countDistinct as countDistinct,
    count_distinct as count_distinct,
    count_if as count_if,
    covar_pop as covar_pop,
    covar_samp as covar_samp,
    create_map as create_map,
    cume_dist as cume_dist,
    current_date as current_date,
    current_timestamp as current_timestamp,
    current_user as current_user,
    date_add as date_add,
    date_diff as date_diff,
    date_format as date_format,
    date_sub as date_sub,
    date_trunc as date_trunc,
    dateadd as dateadd,
    datediff as datediff,
    dayofmonth as dayofmonth,
    dayofweek as dayofweek,
    dayofyear as dayofyear,
    degrees as degrees,
    dense_rank as dense_rank,
    desc as desc,
    desc_nulls_first as desc_nulls_first,
    desc_nulls_last as desc_nulls_last,
    encode as encode,
    exp as exp,
    explode as explode,
    expr as expr,
    extract as extract,
    flatten as flatten,
    floor as floor,
    from_unixtime as from_unixtime,
    get_json_object as get_json_object,
    greatest as greatest,
    grouping_id as grouping_id,
    hash as hash,
    hex as hex,
    hour as hour,
    ifnull as ifnull,
    input_file_name as input_file_name,
    instr as instr,
    isnan as isnan,
    lag as lag,
    last as last,
    lcase as lcase,
    lead as lead,
    least as least,
    left as left,
    length as length,
    levenshtein as levenshtein,
    lit as lit,
    ln as ln,
    locate as locate,
    log as log,
    log10 as log10,
    log2 as log2,
    lower as lower,
    lpad as lpad,
    ltrim as ltrim,
    make_date as make_date,
    map_from_arrays as map_from_arrays,
    max as max,
    max_by as max_by,
    md5 as md5,
    mean as mean,
    min as min,
    min_by as min_by,
    minute as minute,
    month as month,
    months_between as months_between,
    now as now,
    nth_value as nth_value,
    ntile as ntile,
    nullif as nullif,
    nvl as nvl,
    nvl2 as nvl2,
    percent_rank as percent_rank,
    percentile as percentile,
    position as position,
    pow as pow,
    power as power,
    quarter as quarter,
    radians as radians,
    rank as rank,
    regexp_extract as regexp_extract,
    regexp_like as regexp_like,
    regexp_replace as regexp_replace,
    repeat as repeat,
    reverse as reverse,
    right as right,
    rlike as rlike,
    round as round,
    row_number as row_number,
    rpad as rpad,
    rtrim as rtrim,
    second as second,
    shiftLeft as shiftLeft,
    shiftRight as shiftRight,
    shiftleft as shiftleft,
    shiftright as shiftright,
    sign as sign,
    signum as signum,
    sin as sin,
    size as size,
    skewness as skewness,
    sort_array as sort_array,
    soundex as soundex,
    sqrt as sqrt,
    startswith as startswith,
    stddev as stddev,
    stddev_pop as stddev_pop,
    stddev_samp as stddev_samp,
    struct as struct,
    substring as substring,
    sum as sum,
    sumDistinct as sumDistinct,
    sum_distinct as sum_distinct,
    tan as tan,
    timestamp_seconds as timestamp_seconds,
    toDegrees as toDegrees,
    toRadians as toRadians,
    to_date as to_date,
    to_timestamp as to_timestamp,
    translate as translate,
    trim as trim,
    trunc as trunc,
    typeof as typeof,
    ucase as ucase,
    unbase64 as unbase64,
    unhex as unhex,
    unix_date as unix_date,
    unix_timestamp as unix_timestamp,
    upper as upper,
    user as user,
    var_pop as var_pop,
    var_samp as var_samp,
    variance as variance,
    weekofyear as weekofyear,
    when as when,
    year as year,
)



import logging
import sys
import typing as t

from sqlglot import exp

from sqlframe.base.catalog import Column
from sqlframe.base.dataframe import (
    GROUP_DATA,
    NA,
    SESSION,
    STAT,
    WRITER,
    _BaseDataFrame,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


logger = logging.getLogger(__name__)


class NoCachePersistSupportMixin(_BaseDataFrame, t.Generic[SESSION, WRITER, NA, STAT, GROUP_DATA]):
    def cache(self) -> Self:
        logger.warning("This engine does not support caching. Ignoring cache() call.")
        return self

    def persist(self) -> Self:
        logger.warning("This engine does not support persist. Ignoring persist() call.")
        return self


class TypedColumnsFromTempViewMixin(
    _BaseDataFrame, t.Generic[SESSION, WRITER, NA, STAT, GROUP_DATA]
):
    @property
    def _typed_columns(self) -> t.List[Column]:
        table = exp.to_table(self.session._random_id)
        self.session._collect(
            exp.Create(
                this=table,
                kind="VIEW",
                replace=True,
                properties=exp.Properties(expressions=[exp.TemporaryProperty()]),
                expression=self.expression,
            )
        )

        return self.session.catalog.listColumns(
            table.sql(dialect=self.session.input_dialect), include_temp=True
        )



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import fnmatch
import typing as t

from sqlglot import exp, parse_one

from sqlframe.base.catalog import Function, _BaseCatalog
from sqlframe.base.mixins.catalog_mixins import (
    GetCurrentCatalogFromFunctionMixin,
    GetCurrentDatabaseFromFunctionMixin,
    ListCatalogsFromInfoSchemaMixin,
    ListColumnsFromInfoSchemaMixin,
    ListDatabasesFromInfoSchemaMixin,
    ListTablesFromInfoSchemaMixin,
    SetCurrentDatabaseFromSearchPathMixin,
)
from sqlframe.base.util import normalize_string, schema_, to_schema

if t.TYPE_CHECKING:
    from sqlframe.redshift.dataframe import RedshiftDataFrame
    from sqlframe.redshift.session import RedshiftSession


class RedshiftCatalog(
    GetCurrentCatalogFromFunctionMixin["RedshiftSession", "RedshiftDataFrame"],
    GetCurrentDatabaseFromFunctionMixin["RedshiftSession", "RedshiftDataFrame"],
    ListDatabasesFromInfoSchemaMixin["RedshiftSession", "RedshiftDataFrame"],
    ListCatalogsFromInfoSchemaMixin["RedshiftSession", "RedshiftDataFrame"],
    SetCurrentDatabaseFromSearchPathMixin["RedshiftSession", "RedshiftDataFrame"],
    ListTablesFromInfoSchemaMixin["RedshiftSession", "RedshiftDataFrame"],
    ListColumnsFromInfoSchemaMixin["RedshiftSession", "RedshiftDataFrame"],
    _BaseCatalog["RedshiftSession", "RedshiftDataFrame"],
):
    CURRENT_CATALOG_EXPRESSION: exp.Expression = exp.func("current_database")

    def listFunctions(
        self, dbName: t.Optional[str] = None, pattern: t.Optional[str] = None
    ) -> t.List[Function]:
        """
        Returns a t.List of functions registered in the specified database.

        .. versionadded:: 3.4.0

        Parameters
        ----------
        dbName : str
            name of the database to t.List the functions.
            ``dbName`` can be qualified with catalog name.
        pattern : str
            The pattern that the function name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Function`.

        Notes
        -----
        If no database is specified, the current database and catalog
        are used. This API includes all temporary functions.

        Examples
        --------
        >>> spark.catalog.t.listFunctions()
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="to_*")
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="*not_existing_func*")
        []
        """
        if dbName is None:
            schema = schema_(
                db=exp.parse_identifier(
                    self.currentDatabase(), dialect=self.session.output_dialect
                ),
                catalog=exp.parse_identifier(
                    self.currentCatalog(), dialect=self.session.output_dialect
                ),
            )
        else:
            dbName = normalize_string(dbName, from_dialect="input", is_schema=True)
            schema = to_schema(dbName, dialect=self.session.input_dialect)
            if not schema.catalog:
                schema.set(
                    "catalog",
                    exp.parse_identifier(
                        self.currentCatalog(), dialect=self.session.output_dialect
                    ),
                )
        query = parse_one(
            f"""SELECT database_name as catalog, schema_name as namespace, function_name as name
FROM svv_redshift_functions
WHERE database_name = '{schema.catalog}'
and schema_name = '{schema.db}'
ORDER BY function_name;
        """,
            dialect="redshift",
        )
        functions = [
            Function(
                name=normalize_string(x["name"], from_dialect="execution", to_dialect="output"),
                catalog=normalize_string(
                    x["catalog"], from_dialect="execution", to_dialect="output"
                ),
                namespace=[
                    normalize_string(x["namespace"], from_dialect="execution", to_dialect="output")
                ],
                description=None,
                className="",
                isTemporary=False,
            )
            for x in self.session._collect(query)
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            functions = [x for x in functions if fnmatch.fnmatch(x.name, normalized_pattern)]
        return functions

    # def get_columns(self, table_name: t.Union[exp.Table, str]) -> t.Dict[str, exp.DataType]:
    #     table = self.ensure_table(table_name)
    #     if not table.catalog:
    #         table.set("catalog", exp.parse_identifier(self.currentCatalog()))
    #     if not table.db:
    #         table.set("db", exp.parse_identifier(self.currentDatabase()))
    #     sql = f"SHOW COLUMNS FROM TABLE {table.sql(dialect=self.session.input_dialect)}"
    #     results = sorted(self.session._fetch_rows(sql), key=lambda x: x["ordinal_position"])
    #     return {
    #         row["column_name"]: exp.DataType.build(
    #             row["data_type"], dialect=self.session.input_dialect, udt=True
    #         )
    #         for row in results
    #     }



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.group import _BaseGroupedData

if t.TYPE_CHECKING:
    from sqlframe.snowflake.dataframe import SnowflakeDataFrame


class SnowflakeGroupedData(_BaseGroupedData["SnowflakeDataFrame"]):
    pass



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.postgres.session import PostgresSession


class PostgresUDFRegistration(_BaseUDFRegistration["PostgresSession"]): ...



from __future__ import annotations

import typing as t
import warnings

from sqlframe.base.session import _BaseSession
from sqlframe.redshift.catalog import RedshiftCatalog
from sqlframe.redshift.dataframe import RedshiftDataFrame
from sqlframe.redshift.readwriter import (
    RedshiftDataFrameReader,
    RedshiftDataFrameWriter,
)
from sqlframe.redshift.udf import RedshiftUDFRegistration

if t.TYPE_CHECKING:
    from redshift_connector.core import Connection as RedshiftConnection
else:
    RedshiftConnection = t.Any


class RedshiftSession(
    _BaseSession[  # type: ignore
        RedshiftCatalog,
        RedshiftDataFrameReader,
        RedshiftDataFrameWriter,
        RedshiftDataFrame,
        RedshiftConnection,
        RedshiftUDFRegistration,
    ],
):
    _catalog = RedshiftCatalog
    _reader = RedshiftDataFrameReader
    _writer = RedshiftDataFrameWriter
    _df = RedshiftDataFrame
    _udf_registration = RedshiftUDFRegistration

    def __init__(self, conn: t.Optional[RedshiftConnection] = None):
        warnings.warn(
            "RedshiftSession is still in active development. Functions may not work as expected."
        )
        if not hasattr(self, "_conn"):
            super().__init__(conn)

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "redshift"

        @property
        def session(self) -> RedshiftSession:
            return RedshiftSession(**self._session_kwargs)

        def getOrCreate(self) -> RedshiftSession:
            self._set_session_properties()
            return self.session

    builder = Builder()



import typing as t

import pytest
from sqlglot import exp, parse_one
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers

from sqlframe.base import types
from sqlframe.base.util import quote_preserving_alias_or_name, sqlglot_to_spark


@pytest.mark.parametrize(
    "expression, expected",
    [
        ("a", "a"),
        ("a AS b", "b"),
        ("`a`", "`a`"),
        ("`a` AS b", "b"),
        ("`a` AS `b`", "`b`"),
        ("`aB`", "`aB`"),
        ("`aB` AS c", "c"),
        ("`aB` AS `c`", "`c`"),
        ("`aB` AS `Cd`", "`Cd`"),
        # We assume inputs have been normalized so `Cd` is returned as is instead of normalized `cd`
        ("`aB` AS Cd", "Cd"),
    ],
)
def test_quote_preserving_alias_or_name(expression: t.Union[exp.Column, exp.Alias], expected: str):
    assert quote_preserving_alias_or_name(parse_one(expression, dialect="bigquery")) == expected  # type: ignore


@pytest.mark.parametrize(
    "dtype, expected",
    [
        ("STRING", types.StringType()),
        ("VARCHAR(100)", types.VarcharType(100)),
        ("CHAR(100)", types.CharType(100)),
        ("DECIMAL(10, 2)", types.DecimalType(10, 2)),
        ("STRING", types.StringType()),
        ("INTEGER", types.IntegerType()),
        ("BIGINT", types.LongType()),
        ("SMALLINT", types.ShortType()),
        ("FLOAT", types.FloatType()),
        ("DOUBLE", types.DoubleType()),
        ("BOOLEAN", types.BooleanType()),
        ("TIMESTAMP", types.TimestampType()),
        ("DATE", types.DateType()),
        ("DECIMAL", types.DecimalType()),
        ("BINARY", types.BinaryType()),
        ("ARRAY<STRING>", types.ArrayType(types.StringType())),
        ("MAP<STRING, INTEGER>", types.MapType(types.StringType(), types.IntegerType())),
        (
            "STRUCT<a STRING, b INTEGER>",
            types.StructType(
                [
                    types.StructField("a", types.StringType()),
                    types.StructField("b", types.IntegerType()),
                ]
            ),
        ),
        (
            "ARRAY<STRUCT<a STRING, b INTEGER>>",
            types.ArrayType(
                types.StructType(
                    [
                        types.StructField("a", types.StringType()),
                        types.StructField("b", types.IntegerType()),
                    ]
                )
            ),
        ),
    ],
)
def test_sqlglot_to_spark(dtype: str, expected: types.DataType):
    assert sqlglot_to_spark(exp.DataType.build(dtype)) == expected



from __future__ import annotations

import typing as t

from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.standalone.group import StandaloneGroupedData

if t.TYPE_CHECKING:
    from sqlframe.standalone.readwriter import StandaloneDataFrameWriter
    from sqlframe.standalone.session import StandaloneSession


class StandaloneDataFrameNaFunctions(_BaseDataFrameNaFunctions["StandaloneDataFrame"]):
    pass


class StandaloneDataFrameStatFunctions(_BaseDataFrameStatFunctions["StandaloneDataFrame"]):
    pass


class StandaloneDataFrame(
    _BaseDataFrame[
        "StandaloneSession",
        "StandaloneDataFrameWriter",
        "StandaloneDataFrameNaFunctions",
        "StandaloneDataFrameStatFunctions",
        "StandaloneGroupedData",
    ]
):
    _na = StandaloneDataFrameNaFunctions
    _stat = StandaloneDataFrameStatFunctions
    _group_data = StandaloneGroupedData
